package polyglot.ext.predj.ast;

import polyglot.ast.*;
import polyglot.ext.jl.ast.*;
import polyglot.ext.predj.types.*;
import polyglot.visit.*;
import polyglot.ext.predj.visit.*;
import polyglot.types.*;

import java.util.*;

/** An AST node representing a dispatcher for a set of dispatcher mates. **/
public class PredJDispatcher_c extends Term_c implements PredJDispatcher {

	// the list of mates (PredJMethodDecls)
	// that this dispatcher dispatches to
    protected List dispatcherMates;

	// does this dispatcher include an abstract method, which
	// the other mates partially override?  that is the way
	// we express 'partially abstract' gfs
    protected boolean hasAbstractMethod = false;

    protected PredJDispatcherInstance di;

    public PredJDispatcher_c(List mates, PredJTypeSystem ts) {
	super(null);
	dispatcherMates = mates;

	    // give each mate a unique number and make the dispatcher
	    // instance.  also set a flag telling each method that it has
	    // mates (i.e., is part of a dispatcher).
	    // also check for and remove abstract methods,
	    // which are used when defining partially abstract gfs.
	    // if we find one, set the corresponding flag.
	Iterator i = dispatcherMates.iterator();
	int count = 0;
	List mis = new LinkedList();
	while(i.hasNext()) {
	    PredJMethodDecl md = (PredJMethodDecl) i.next();
	    md.setHasMates();
	    if (md.flags().isAbstract()) {
		hasAbstractMethod = true;
		i.remove();
	    } else {
		md.setMateNumber(count);
		mis.add(md.methodInstance());
		count++;
	    }
	}
	di = ts.dispatcherInstance(mis, hasAbstractMethod);
    }

    public List dispatcherMates() {
	return dispatcherMates;
    }

    public List[] partialOrder() {
	return di.partialOrder();
    }


    public Node implementationSideTypeCheck(
	ImplementationSideTypeChecker itc) throws SemanticException {
	    // implementation-side typechecking:  make sure this
	    // generic function is complete (for the receiver containing
	    // these mates) and that the mates are not ambiguous with one
	    // another.  by our encapsulated-style dispatching semantics,
	    // the mates cannot be ambiguous with other methods of the gf.

	PredJTypeSystem ts = (PredJTypeSystem) itc.typeSystem();
	ClassType ct = itc.context().currentClass();

	boolean complete = di.isComplete(ct, ts);
	if (!hasAbstractMethod && !complete) {
	    // if the dispatcher does not have an abstract method,
	    // then we require it to be complete, even if it's in an
	    // abstract class.  this prevents accidentally not making
	    // a gf complete when it was meant to be.
	    String suggestion = "This method's associated operation is " +
		"not fully implemented. " +
		"One way to resolve the problem is to add "
		+ "a default method, which is a method that has no " +
		"\'when\' clause.";
	    if (ct.flags().isAbstract()) {
		suggestion += " Or, since this class is abstract, you " +
		    "may instead declare an abstract method with no " +
		    "\'when\' clause, which the current methods partially "
		    + "override; subclasses must then provide the rest.";
	    }
	    throw new SemanticException(suggestion,
		((MethodDecl) dispatcherMates().get(0)).position());
	} else if (hasAbstractMethod && complete) {
	    // if the dispatcher has an abstract method,
	    // then we require it to be incomplete, as a sanity check.
	    throw new SemanticException(
		"This method's associated operation is fully " +
		"implemented. Therefore, there should not also exist " +
		"an abstract method for the operation.",
		((MethodDecl) dispatcherMates().get(0)).position());
	}
	
	    // always check ambiguities; these are ticking time bombs,
	    // so we might as well check them early.
	di.checkUnambiguity(ts);

	di.checkSatisfiablePredicates(ts);

	return this;
    }


/* These methods are necessary because PredJDispatcher_c recursively implements
   the Term interface, which is the interface for nodes that can have dataflow
   analysis performed on them.  We implement dummy versions of these methods,
   which simply skip over this node for the purposes of dataflow analysis.
   That probably makes sense anyway, since dispatchers are really synthetic
   class members. No guarantee that DFA works properly with this
   implementation, however. */
    
    /**
     * Return the first (sub)term performed when evaluating this
     * term.
     */
    public Term entry() {
	return this;
    }

    /**
     * Visit this node, calling calling v.edge() for each successor in succs,
     * if data flows on that edge.
     */
    public List acceptCFG(CFGBuilder v, List succs) {
	return succs;
    }
}
