
package polyglot.ext.predj.visit;

import polyglot.visit.NodeVisitor;
import polyglot.ast.Node;
import polyglot.ext.jl.ast.*;
import polyglot.ext.predj.ast.*;
import polyglot.ast.Binary;
import polyglot.ast.Unary;

// a pass that determines the "polarity" of a subpredicate in a method's
// predicate.  if the polarity is true, that means that whenever the
// overall predicate is true, then so is this subexpression.  in that case,
// more refined types (and new bindings, if any) produced by that
// subpredicate may safely be propagated for use in the method's body.

// the current algorithm only allows refined types to propagate from within
// any number of ANDs; ORs and NOTs make the polarity false, and it never
// can become true again.  it may be possible to relax this eventually;
// two NOTs revert the polarity, and we can do a GLB computation to allow
// refinements to propagate that are defined on both sides of an OR (see
// the appendix of the GUD paper).

public class PredicatePolarity extends NodeVisitor {

    private boolean polarity;

    public PredicatePolarity() {
	this(true);
    }

    public PredicatePolarity(boolean polarity) {
	super();
	this.polarity = polarity;
    }

    public NodeVisitor enter(Node n) {
	if (n instanceof Binary) {
	    Binary b = (Binary) n;
	    if (b.operator().equals(Binary.COND_OR)) {
		return new PredicatePolarity(false);
	    } else {
		return super.enter(n);
	    }
	} else if (n instanceof Unary) {
	    Unary u = (Unary) n;
	    if (u.operator().equals(Unary.NOT)) {
		return new PredicatePolarity(false);
	    } else {
		return super.enter(n);
	    }
	} else
	    return super.enter(n);
    }

	// pass polarity info on to specializers and bindings
    
    public Node leave(Node old, Node n, NodeVisitor v) {
	if (n instanceof BindingOrSpecializerNode) {
	    return ((BindingOrSpecializerNode) n).polarity(this.polarity);
	}
	return super.leave(old, n, v);
    }
}
