package polyglot.ext.predj.visit;

import polyglot.visit.*;
import polyglot.ast.*;
import polyglot.ext.jl.ast.*;
import polyglot.ext.predj.ast.*;
import polyglot.types.*;
import polyglot.ext.predj.types.*;
import polyglot.util.*;
import polyglot.frontend.Job;

import java.util.*;

/** Visitor that makes sure that pure methods are actually pure, and that
    only pure methods are called in predicates.  **/
public class PurityChecker extends ContextVisitor {
    protected boolean insidePredicate;

    public PurityChecker(Job job, TypeSystem ts, NodeFactory nf) {
	super(job, ts, nf);
        insidePredicate = false;
    }

    protected boolean insidePureMethod() {
        return context().currentCode().flags().contains(
                                                  ((PredJTypeSystem)ts).Pure());
    }

    protected boolean isPureMethod(MethodInstance mi) {
        return mi.flags().contains(((PredJTypeSystem)ts).Pure());
    }

    public NodeVisitor enterCall(Node p, Node n)
        throws SemanticException {
        if (p instanceof PredicateNode) {
	    insidePredicate = true;
	    return this;
	}
	return super.enterCall(p, n);
    }

    protected Node leaveCall(Node n) throws SemanticException
    {
	if (n instanceof PredicateNode) {
	    insidePredicate = false;
	    return n;
	} else if (n instanceof MethodDecl) {
	    MethodDecl md = (MethodDecl) n;
	    if (!isPureMethod(md.methodInstance())) {
		List overrides = md.methodInstance().overrides();
		if (overrides.size() > 1 &&
		    isPureMethod((MethodInstance)overrides.get(1)))
		    throw new SemanticException(
			"This method must be declared pure.");
	    }
	    return md;
	} else if (n instanceof Call) {
	    Call node = (Call) n;
	    MethodInstance mi = node.methodInstance();
	    if ((!isPureMethod(mi)) && (insidePureMethod() || insidePredicate))
		throw new SemanticException(
		    "Non-pure calls are not allowed here.");
	    return node;
	} else if (n instanceof ArrayAccessAssign) {
	    if (insidePureMethod())
		throw new SemanticException(
		    "Pure methods cannot modify arrays.");
	    return n;
	} else if (n instanceof FieldAssign) {
	    if (insidePureMethod())
		throw new SemanticException(
		    "Pure methods cannot modify fields.");
	    return n;
	} else if (n instanceof New) {
	    if (insidePureMethod())
		throw new SemanticException(
		    "Pure methods cannot use constructors.");
	    return n;
	} else if (n instanceof Unary) {
	    Unary node = (Unary) n;
	    if (insidePureMethod() && !(node.expr() instanceof Local)) {
		Unary.Operator op = node.operator();
		if (op == Unary.PRE_DEC || op == Unary.POST_DEC ||
		    op == Unary.PRE_INC || op == Unary.POST_INC)
		    throw new SemanticException("Pure Methods cannot modify "+
						"non-local variables.");
	    }
	    return node;
	} else
	    return super.leaveCall(n);
    }
}
